@@ -260,7 +260,7 @@ impl CalloutScriptCache {
         }
     }
 
-    fn lookup_callout_script(&self, parent: &str, mdev_type: &str) -> Option<CalloutScriptInfo> {
+    fn lookup_cached_script(&self, parent: &str, mdev_type: &str) -> Option<CalloutScriptInfo> {
         self.callouts
             .iter()
             .find(|cs| cs.mdev_type == mdev_type && cs.parent == parent)
@@ -270,7 +270,16 @@ impl CalloutScriptCache {
     pub fn find_versioned_script(&mut self, dev: &MDev) -> Option<CalloutScriptInfo> {
         // check already found scripts
         let mut dev = dev.clone();
-        let mut callout = callout(&mut dev);
+        let mut callout = match callout(&mut dev) {
+            Ok(c) => c,
+            Err(e) => {
+                debug!(
+                    "cannot find a callout script for mdev with uuid {:?}: {:?}",
+                    dev.uuid, e
+                );
+                return None;
+            }
+        };
         let mdev_type = match callout.dev.mdev_type() {
             Ok(t) => t.clone(),
             Err(_) => {
@@ -286,7 +295,7 @@ impl CalloutScriptCache {
             }
         };
         debug!("Looking up callout script for mdev type '{:?}'", mdev_type);
-        match self.lookup_callout_script(&parent, &mdev_type) {
+        match self.lookup_cached_script(&parent, &mdev_type) {
             Some(cs) => {
                 debug!(
                     "Callout script lookup for mdev type '{:?}' and parent {:?} successful",
@@ -343,7 +352,7 @@ impl CalloutScriptCache {
 
 pub trait CheckProcessOutput {
     fn check(&self, p: PathBuf, o: Output) -> Result<(PathBuf, Output)>;
-    fn process(&self, c: &mut Callout<'_, '_>, p: PathBuf, o: Output) -> Result<Option<Output>>;
+    fn process(&self, c: &mut Callout<'_>, p: PathBuf, o: Output) -> Result<Option<Output>>;
 }
 
 struct DefaultCheckProcessOutput;
@@ -353,7 +362,7 @@ impl CheckProcessOutput for DefaultCheckProcessOutput {
         Ok((p, o))
     }
 
-    fn process(&self, c: &mut Callout<'_, '_>, p: PathBuf, o: Output) -> Result<Option<Output>> {
+    fn process(&self, c: &mut Callout<'_>, p: PathBuf, o: Output) -> Result<Option<Output>> {
         c.print_err(&o, &p);
         match o.status.code() {
             Some(0) => {
@@ -384,7 +393,7 @@ impl CheckProcessOutput for CapabilitiesCheckProcessOutput {
         }
     }
 
-    fn process(&self, c: &mut Callout<'_, '_>, p: PathBuf, o: Output) -> Result<Option<Output>> {
+    fn process(&self, c: &mut Callout<'_>, p: PathBuf, o: Output) -> Result<Option<Output>> {
         c.print_err(&o, &p);
         match CalloutScriptCache::parse_script_capabilities(&o) {
             Some(cv) => {
@@ -408,26 +417,26 @@ impl CheckProcessOutput for CapabilitiesCheckProcessOutput {
     }
 }
 
-pub struct Callout<'a, 'b> {
+pub struct Callout<'a> {
     state: State,
     script: Option<CalloutScriptInfo>,
-    pub dev: &'b mut MDev<'a>,
+    pub dev: &'a mut MDev,
 }
 
-pub fn callout<'a, 'b>(dev: &'b mut MDev<'a>) -> Callout<'a, 'b> {
+pub fn callout(dev: &mut MDev) -> Result<Callout> {
     Callout::new(dev)
 }
 
-impl<'a, 'b> Callout<'a, 'b> {
-    pub fn new(dev: &'b mut MDev<'a>) -> Callout<'a, 'b> {
+impl<'a> Callout<'a> {
+    pub fn new(dev: &'a mut MDev) -> Result<Callout<'a>> {
         if dev.mdev_type.is_none() {
-            panic!("Device dev must have a defined mdev_type!")
+            return Err(anyhow!("Device must have a defined mdev_type"));
         }
-        Callout {
+        Ok(Callout {
             state: State::None,
             script: None,
             dev,
-        }
+        })
     }
 
     fn find_callout_script(&self) -> Option<CalloutScriptInfo> {
@@ -440,19 +449,19 @@ impl<'a, 'b> Callout<'a, 'b> {
             // live is only supported when script with versioning exists
             debug!("No callout script with version support found that supports live modify");
             return Err(anyhow!(
-                "No callout script with version support found that supports live modify"
+                "Live modification is not supported for this device type. Changing the device configuration requires stopping and restarting the device."
             ));
         }
 
         let mut res = Ok(());
-        let mut existing = MDev::new(self.dev.env, self.dev.uuid);
-        if existing.load_from_sysfs().is_ok() && existing.active {
-            if existing.parent != self.dev.parent {
+        let sysfs_data = MDevSysfsData::load_for_mdev(self.dev)?;
+        if let Some(sysfs_data) = sysfs_data {
+            if Some(sysfs_data.parent) != self.dev.parent {
                 debug!("Device exists under different parent - cannot run live update");
                 res = Err(anyhow!(
                     "Device exists under different parent - cannot run live update"
                 ));
-            } else if existing.mdev_type != self.dev.mdev_type {
+            } else if Some(sysfs_data.mdev_type) != self.dev.mdev_type {
                 debug!("Device exists with different type - cannot run live update");
                 res = Err(anyhow!(
                     "Device exists with different type - cannot run live update"
