@@ -3,10 +3,10 @@
 use crate::environment::Environment;
 use anyhow::{anyhow, Context, Result};
 use log::{debug, warn};
-use std::convert::TryInto;
 use std::fs;
-use std::io::Read;
+use std::io::{ErrorKind, Read};
 use std::path::{Path, PathBuf};
+use std::rc::Rc;
 use std::vec::Vec;
 use uuid::Uuid;
 
@@ -16,20 +16,99 @@ pub enum FormatType {
     Defined,
 }
 
+pub struct MDevSysfsData {
+    pub uuid: Uuid,
+    pub parent: String,
+    pub mdev_type: String,
+}
+
+impl MDevSysfsData {
+    pub fn load(env: &Rc<dyn Environment>, uuid: &Uuid) -> Result<Option<MDevSysfsData>> {
+        let active_path = Self::active_path(env.clone(), uuid);
+        let parent = Self::load_parent_from_sysfs(&active_path)
+            .map(Some)
+            .or_else(|e: std::io::Error| match e.kind() {
+                ErrorKind::NotFound => {
+                    debug!("Mdev {:?} does no longer exist in sysfs", uuid);
+                    Ok(None)
+                }
+                _ => Err(e),
+            })?;
+        let mdev_type = Self::load_mdev_type_from_sysfs(&active_path)
+            .map(Some)
+            .or_else(|e| match e.kind() {
+                std::io::ErrorKind::NotFound => {
+                    debug!("Mdev {:?} does no longer exist in sysfs", uuid);
+                    Ok(None)
+                }
+                _ => Err(e),
+            })?;
+        if let (Some(parent), Some(mdev_type)) = (parent, mdev_type) {
+            Ok(Some(MDevSysfsData {
+                uuid: uuid.to_owned(),
+                parent,
+                mdev_type,
+            }))
+        } else {
+            debug!("Mdev {:?} does not exist in sysfs", uuid);
+            Ok(None)
+        }
+    }
+
+    pub fn load_for_mdev(mdev: &MDev) -> Result<Option<MDevSysfsData>> {
+        Self::load(&mdev.env, &mdev.uuid)
+    }
+
+    fn active_path(env: Rc<dyn Environment>, uuid: &Uuid) -> PathBuf {
+        env.mdev_base().join(uuid.hyphenated().to_string())
+    }
+
+    fn load_parent_from_sysfs<P: AsRef<Path>>(active_path: P) -> std::io::Result<String> {
+        let canonpath = fs::canonicalize(&active_path)?;
+        let sysfsparent = canonpath.parent().ok_or_else(|| {
+            std::io::Error::new(
+                ErrorKind::InvalidInput,
+                anyhow!("Path to parent of mdev {:?} does not exist", canonpath),
+            )
+        })?;
+        Self::canonical_basename(sysfsparent)
+    }
+
+    fn load_mdev_type_from_sysfs<P: Into<PathBuf>>(active_path: P) -> std::io::Result<String> {
+        let mut typepath: PathBuf = active_path.into();
+        typepath.push("mdev_type");
+        Self::canonical_basename(typepath)
+    }
+
+    fn canonical_basename<P: AsRef<Path>>(path: P) -> std::io::Result<String> {
+        let path = fs::canonicalize(path)?;
+        let fname = path
+            .file_name()
+            .ok_or_else(|| std::io::Error::new(ErrorKind::InvalidInput, anyhow!("Invalid path")))?;
+        match fname.to_str() {
+            Some(x) => Ok(x.to_string()),
+            None => Err(std::io::Error::new(
+                ErrorKind::InvalidInput,
+                anyhow!("Invalid file name"),
+            )),
+        }
+    }
+}
+
 /// Representation of a mediated device
 #[derive(Debug, Clone)]
-pub struct MDev<'a> {
+pub struct MDev {
     pub uuid: Uuid,
     pub active: bool,
     pub autostart: bool,
     pub parent: Option<String>,
     pub mdev_type: Option<String>,
     pub attrs: Vec<(String, String)>,
-    pub env: &'a dyn Environment,
+    pub env: Rc<dyn Environment>,
 }
 
-impl<'a> MDev<'a> {
-    pub fn new(env: &'a dyn Environment, uuid: Uuid) -> MDev<'a> {
+impl MDev {
+    pub fn new(env: Rc<dyn Environment>, uuid: Uuid) -> MDev {
         MDev {
             uuid,
             active: false,
@@ -41,9 +120,26 @@ impl<'a> MDev<'a> {
         }
     }
 
-    pub fn path(&self) -> PathBuf {
+    pub fn new_from_jsonfile(
+        env: Rc<dyn Environment>,
+        uuid: Uuid,
+        parent: String,
+        jsonfile: PathBuf,
+    ) -> Result<Self> {
+        let _ = std::fs::File::open(&jsonfile)
+            .with_context(|| format!("Unable to read file {:?}", jsonfile))?;
+        let filecontents = fs::read_to_string(&jsonfile)
+            .with_context(|| format!("Unable to read jsonfile {:?}", jsonfile))?;
+        let jsonval = serde_json::from_str(&filecontents)?;
+
+        let mut d = MDev::new(env, uuid);
+        d.load_from_json(parent, &jsonval)?;
+        Ok(d)
+    }
+
+    pub fn active_path(&self) -> PathBuf {
         let mut p = self.env.mdev_base();
-        p.push(self.uuid.to_hyphenated().to_string());
+        p.push(self.uuid.hyphenated().to_string());
         p
     }
 
@@ -52,7 +148,7 @@ impl<'a> MDev<'a> {
         self.parent.as_ref().ok_or_else(|| {
             anyhow!(
                 "Device {} is missing a parent",
-                self.uuid.to_hyphenated().to_string()
+                self.uuid.hyphenated().to_string()
             )
         })
     }
@@ -62,65 +158,65 @@ impl<'a> MDev<'a> {
         self.mdev_type.as_ref().ok_or_else(|| {
             anyhow!(
                 "Device {} is missing a mdev_type",
-                self.uuid.to_hyphenated().to_string()
+                self.uuid.hyphenated().to_string()
             )
         })
     }
 
-    pub fn persist_path(&self) -> Option<PathBuf> {
+    pub fn persistent_path(&self) -> Option<PathBuf> {
         self.parent.as_ref().map(|x| {
-            let mut path = self.env.persist_base();
+            let mut path = self.env.config_base();
             path.push(x);
-            path.push(self.uuid.to_hyphenated().to_string());
+            path.push(self.uuid.hyphenated().to_string());
             path
         })
     }
 
     pub fn is_defined(&self) -> bool {
-        match self.persist_path() {
+        match self.persistent_path() {
             Some(p) => p.exists(),
             None => false,
         }
     }
 
-    pub fn load_from_sysfs(&mut self) -> Result<()> {
-        debug!("Loading device '{:?}' from sysfs", self.uuid);
-        if !self.path().exists() {
-            debug!("loaded device {:?}", self);
-            return Ok(());
+    pub fn set_sysfs_data(&mut self, sysfs_data: Option<MDevSysfsData>) {
+        if let Some(d) = sysfs_data {
+            if self.uuid != d.uuid {
+                warn!(
+                    "Attempting to set sysfs data for device {} from sysfs data for UUID {}",
+                    self.uuid, d.uuid
+                );
+                return;
+            }
+            self.parent = Some(d.parent);
+            self.mdev_type = Some(d.mdev_type);
+            self.active = true;
+        } else {
+            self.active = false;
         }
+    }
 
-        let canonpath = self.path().canonicalize()?;
-        let sysfsparent = canonpath.parent().unwrap();
-        let parentname = canonical_basename(sysfsparent)?;
-        if self.parent.is_some() && self.parent.as_ref() != Some(&parentname) {
+    pub fn sysfs_data_matches(&self, sysfs_data: &MDevSysfsData) -> bool {
+        if self.parent.as_ref() != Some(&sysfs_data.parent) {
             debug!(
                 "Active mdev {:?} has different parent: {}!={}. No match.",
                 self.uuid,
                 self.parent.as_ref().unwrap(),
-                parentname
+                sysfs_data.parent
             );
-            return Ok(());
+            return false;
         }
-        let mut typepath = self.path();
-        typepath.push("mdev_type");
-        let mdev_type = canonical_basename(typepath)?;
-        if self.mdev_type.is_some() && self.mdev_type.as_ref() != Some(&mdev_type) {
+
+        if self.mdev_type.as_ref() != Some(&sysfs_data.mdev_type) {
             debug!(
                 "Active mdev {:?} has different type: {}!={}. No match.",
                 self.uuid,
                 self.mdev_type.as_ref().unwrap(),
-                mdev_type
+                sysfs_data.mdev_type
             );
-            return Ok(());
+            return false;
         }
-
-        // active device in sysfs matches this device. update information
-        self.mdev_type = Some(mdev_type);
-        self.parent = Some(parentname);
-        self.active = true;
-        debug!("loaded device {:?}", self);
-        Ok(())
+        true
     }
 
     pub fn add_attributes(&mut self, attrs: &serde_json::Value) -> Result<()> {
@@ -142,7 +238,9 @@ impl<'a> MDev<'a> {
                     }
                     // get the key and value from the first (only) map entry
                     if let Some((key, val)) = attrobj.iter().next() {
-                        let valstr = val.as_str().unwrap();
+                        let valstr = val.as_str().ok_or_else(|| {
+                            anyhow!("invalid JSON format for attribute {{{:?}, {}}}: value must be of type str", key, val)
+                        })?;
                         self.attrs.push((key.to_string(), valstr.to_string()));
                     }
                 }
@@ -190,7 +288,7 @@ impl<'a> MDev<'a> {
 
     // load the stored definition from disk if it exists
     pub fn load_definition(&mut self) -> Result<()> {
-        if let Some(path) = self.persist_path() {
+        if let Some(path) = self.persistent_path() {
             let mut f = fs::File::open(path)?;
             let mut contents = String::new();
             f.read_to_string(&mut contents)?;
@@ -215,7 +313,7 @@ impl<'a> MDev<'a> {
             }
         }
 
-        let mut output = self.uuid.to_hyphenated().to_string();
+        let mut output = self.uuid.hyphenated().to_string();
         output.push(' ');
         output.push_str(self.parent()?);
         output.push(' ');
@@ -240,14 +338,23 @@ impl<'a> MDev<'a> {
         }
 
         output.push('\n');
-        if verbose && !self.attrs.is_empty() {
+        if verbose {
+            let attr_string = self.fmt_attrs();
+            output.push_str(&attr_string);
+        }
+        Ok(output)
+    }
+
+    fn fmt_attrs(&self) -> String {
+        let mut output = String::new();
+        if !self.attrs.is_empty() {
             output.push_str("  Attrs:\n");
             for (i, (key, value)) in self.attrs.iter().enumerate() {
                 let txtattr = format!("    @{{{}}}: {{\"{}\":\"{}\"}}\n", i, key, value);
                 output.push_str(&txtattr);
             }
         }
-        Ok(output)
+        output
     }
 
     pub fn to_json(&self, include_uuid: bool) -> Result<serde_json::Value> {
@@ -265,7 +372,7 @@ impl<'a> MDev<'a> {
             .collect();
         partial.insert("attrs".to_string(), jsonattrs.into());
 
-        let full = serde_json::json!({ self.uuid.to_hyphenated().to_string(): partial });
+        let full = serde_json::json!({ self.uuid.hyphenated().to_string(): partial });
 
         match include_uuid {
             true => Ok(full),
@@ -275,7 +382,7 @@ impl<'a> MDev<'a> {
 
     pub fn stop(&mut self) -> Result<()> {
         debug!("Removing mdev {:?}", self.uuid);
-        let mut remove_path = self.path();
+        let mut remove_path = self.active_path();
         remove_path.push("remove");
         debug!("remove path '{:?}'", remove_path);
         match fs::write(remove_path, "1") {
@@ -289,7 +396,7 @@ impl<'a> MDev<'a> {
 
     fn find_parent_dir(&self) -> Result<PathBuf> {
         let parent = self.parent()?;
-        let path: PathBuf = self.env.parent_base().join(&parent);
+        let path: PathBuf = self.env.parent_base().join(parent);
 
         if path.is_dir() {
             return Ok(path);
@@ -315,16 +422,23 @@ impl<'a> MDev<'a> {
         debug!("Creating mdev {:?}", self.uuid);
         let parent = self.parent()?;
         let mdev_type = self.mdev_type()?;
-        let mut existing = MDev::new(self.env, self.uuid);
-
-        if existing.load_from_sysfs().is_ok() && existing.active {
-            if existing.parent != self.parent {
-                return Err(anyhow!("Device exists under different parent"));
+        match MDevSysfsData::load(&self.env, &self.uuid) {
+            Ok(Some(mdev_sysfs_data)) => {
+                if Some(mdev_sysfs_data.parent) != self.parent {
+                    return Err(anyhow!("Device exists under different parent"));
+                }
+                if Some(mdev_sysfs_data.mdev_type) != self.mdev_type {
+                    return Err(anyhow!("Device exists with different type"));
+                }
+                return Err(anyhow!("Device already exists"));
             }
-            if existing.mdev_type != self.mdev_type {
-                return Err(anyhow!("Device exists with different type"));
+            Ok(_) => (),
+            Err(e) => {
+                warn!(
+                    "A sysfs lookup for device {} caused the error: {:?}",
+                    self.uuid, e
+                );
             }
-            return Err(anyhow!("Device already exists"));
         }
 
         let mut path = self.find_parent_dir()?;
@@ -336,7 +450,7 @@ impl<'a> MDev<'a> {
                 parent
             ));
         }
-        path.push(&mdev_type);
+        path.push(mdev_type);
         debug!("Checking parent for mdev type {}: {:?}", mdev_type, path);
         if !path.is_dir() {
             return Err(anyhow!(
@@ -360,7 +474,7 @@ impl<'a> MDev<'a> {
         path.pop();
         path.push("create");
         debug!("Creating mediated device: {:?} -> {:?}", self.uuid, path);
-        match fs::write(path, self.uuid.to_hyphenated().to_string()) {
+        match fs::write(path, self.uuid.hyphenated().to_string()) {
             Ok(_) => {
                 self.active = true;
                 Ok(())
@@ -368,7 +482,7 @@ impl<'a> MDev<'a> {
             Err(e) => Err(e).with_context(|| {
                 format!(
                     "Failed to create mdev {}, type {} on {}",
-                    self.uuid.to_hyphenated(),
+                    self.uuid.hyphenated(),
                     mdev_type,
                     parent
                 )
@@ -381,7 +495,7 @@ impl<'a> MDev<'a> {
 
         debug!("Setting attributes for mdev {:?}", self.uuid);
         for (k, v) in self.attrs.iter() {
-            if let Err(e) = write_attr(&self.path(), k, v) {
+            if let Err(e) = write_attr(&self.active_path(), k, v) {
                 self.stop()?;
                 return Err(e);
             }
@@ -392,7 +506,7 @@ impl<'a> MDev<'a> {
 
     pub fn write_config(&self) -> Result<()> {
         let jsonstring = serde_json::to_string_pretty(&self.to_json(false)?)?;
-        let path = self.persist_path().unwrap();
+        let path = self.persistent_path().unwrap();
         let parentdir = path.parent().unwrap();
         debug!("Ensuring parent directory {:?} exists", parentdir);
         fs::create_dir_all(parentdir)?;
@@ -406,23 +520,35 @@ impl<'a> MDev<'a> {
     }
 
     pub fn undefine(&mut self) -> Result<()> {
-        let p = self.persist_path().ok_or_else(|| {
-            anyhow!(
-                "Failed to undefine {}",
-                self.uuid.to_hyphenated().to_string()
-            )
-        })?;
+        let p = self
+            .persistent_path()
+            .ok_or_else(|| anyhow!("Failed to undefine {}", self.uuid.hyphenated().to_string()))?;
 
         fs::remove_file(&p).with_context(|| format!("Failed to remove file {:?}", p))?;
         Ok(())
     }
 
-    pub fn add_attribute(&mut self, name: String, value: String, index: Option<u32>) -> Result<()> {
+    fn attribute_hint(&self) -> String {
+        match self.attrs.is_empty() {
+            true => format!("Device {} has no attributes", self.uuid.hyphenated()),
+            false => self.fmt_attrs(),
+        }
+    }
+
+    pub fn add_attribute(
+        &mut self,
+        name: String,
+        value: String,
+        index: Option<usize>,
+    ) -> Result<()> {
         match index {
             Some(i) => {
-                let i: usize = i.try_into().unwrap();
                 if i > self.attrs.len() {
-                    return Err(anyhow!("Attribute index {} is invalid", i));
+                    return Err(anyhow!(
+                        "Attribute index {} is invalid\n{}",
+                        i,
+                        self.attribute_hint()
+                    ));
                 }
                 self.attrs.insert(i, (name, value));
             }
@@ -432,12 +558,15 @@ impl<'a> MDev<'a> {
         Ok(())
     }
 
-    pub fn delete_attribute(&mut self, index: Option<u32>) -> Result<()> {
+    pub fn delete_attribute(&mut self, index: Option<usize>) -> Result<()> {
         match index {
             Some(i) => {
-                let i: usize = i.try_into().unwrap();
-                if i > self.attrs.len() {
-                    return Err(anyhow!("Attribute index {} is invalid", i));
+                if i >= self.attrs.len() {
+                    return Err(anyhow!(
+                        "Attribute index {} is invalid\n{}",
+                        i,
+                        self.attribute_hint()
+                    ));
                 }
                 self.attrs.remove(i);
             }
@@ -450,19 +579,6 @@ impl<'a> MDev<'a> {
     }
 }
 
-fn canonical_basename<P: AsRef<Path>>(path: P) -> Result<String> {
-    let path = fs::canonicalize(path)?;
-    let fname = path.file_name();
-    if fname.is_none() {
-        return Err(anyhow!("Invalid path"));
-    }
-    let fname = fname.unwrap().to_str();
-    match fname {
-        Some(x) => Ok(x.to_string()),
-        None => Err(anyhow!("Invalid file name")),
-    }
-}
-
 fn write_attr(basepath: &Path, attr: &str, val: &str) -> Result<()> {
     debug!("Writing attribute '{}' -> '{}'", attr, val);
     let path = basepath.join(attr);
